------------------------------------------------------------------------
-- Copyright Notice:                                                    
-- (c) Copyright 2004 - 2008, National Optical Astronomy Observatory, 
-- which is operated for the National Science Foundation by the    
-- Association of Universities for Research in Astronomy under    
-- Cooperative Agreement No. AST-0132798.                                                                                                                           
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   SYSTEM TOP LEVEL  
-- MODULE       :   WISHBONE BUS MASTER INTERFACE
-- VERSION      :   1.06
------------------------------------------------------------------------
-- FILE NAME    :   WBMASTERINTERFACE.VHD
-- AUTHOR       :   PETER MOORE
-- DATE         :   03/19/2010
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.3
------------------------------------------------------------------------
-- DESCRIPTION  :
-- Master bus module connects to a bus structure that supports 32-bit data
-- transactions to take place via an interconnect logic module (INTERCON). 
-- The master interface supports read and write transactions on a standard 
-- Wishbone bus (see http://en.wikipedia.org/wiki/Wishbone_Bus or 
-- http://www.opencores.org/projects.cgi/web/wishbone/wbspec_b3.pdf ). 
-- The bus master module acts as an agent for a 'client module' to perform
-- such transactions with full support for the multi-master architecture of
-- the Torrent firmware design. However, in the Torrent design, we use the 
-- latent bus (i.e. when it is not being used to transport data) as an 
-- asynchronous out of band data path to comunicate the system status to all
-- of the firmware modules in common. In this respect this firmware deviates
-- from the published specification.                  
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- Note that write data requests (WrtDataReq) have a higher priority than
-- read requests (RdDataReq)
-- 
-- Bus Cycle Tag identifiers (TGC_O)
-- TGC_O + WE_I  
--   00 = Normal bus write cycle
--   01 = spare
--   10 = Start exposure - produces a slave start exposure pulse and the required start vector (carried on the address signals).
--   11 = Local module reset - produces a reset strobe at the user slave interface port.
-- TGC_O + /WE_I
--   00 = normal bus read cycle
--   01 = spare
--   10 = spare
--   11 = spare
--
-- Bus Data Tag identifiers (TGD_O, TGD_I) 
-- Identifies the source of the command during a write cycle or the destination of data during a read cycle
--    This is established in the Command Mux source as:
--   constant SRC_FPDP : std_logic_vector(3 downto 0) := "0001";
--   constant SRC_SYNC : std_logic_vector(3 downto 0) := "0010";
--   constant SRC_UART : std_logic_vector(3 downto 0) := "0100";
--   constant SRC_GIGE : std_logic_vector(3 downto 0) := "1000";
--
-- BusEventStatus register bit descriptions
-- BIT significance
--  0 High indicates that slave interface ERR acknowledgements have occured
--  1 High indicates that slave interface RTY acknowledgements have occured
--  2 High indicated that slave interface normal ACK timouts have occured
--  3 High indicates that Master Bus Grant errors have occured
--
--  The three diagnostic registers (BusErrorReg, BusGrantReg, BusTimeOutReg)
--  keep count of the number of errors that occur of each type. 
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	  AUTHOR  VERSION  MODIFICATIONS
-- 06/03/2008  PCM   1.00   ORIGINAL VERSION
-- 10/02/2008  PCM   1.01   1st WORKING VERSION - BETA !
-- 06/01/2009  PCM   1.02   ADDED RESET AND START EXPOSURE CAPABILITY TO THE TGC SIGNALS
-- 02/23/2010  PCM   1.03   ADDED EXPANDED THE READ AND WRITE ADDRESS PORTS TO ACCOMODATE THE 8 SLAVE INTERFACE VIA INTERCON VERSION 1.05
-- 03/19/2010   PCM   1.06  RESOLVED SEVERAL PROBLEMS: STATUS OUTPUT CONTAMINATION, BURST WRITE ARBITRATION AND TIMING.  
------------------------------------------------------------------------
-- PORT DEFINITIONS :
-- Wishbone bus interface signals
-- CLK_I              Wishbone / System clock source input to interface
-- RST_I              Wishbone / system reset signal input to interface
-- CYC_O              Wishbone active cycle request from master to bus arbitration logic
-- CYC_GRNT           Wishbone bus grant signal from arbitration logic
-- STB_O              Individual slave address strobes. From master to slaves.
-- WE_O               Wishbone Write Enable signal. From master to slave.
-- ADR_O              Wishbone read / write address. From master to slave.
-- DAT_O              Wishbone data bus. From master to slave
-- TGD_O              Wishbone data identification tag signal. From master to slave.
-- TGC_O              Wishbone cycle type tag - see implementation notes above.
-- ACK_I              Slave interface normal bus cycle acknowledgement signal. From slave to master.
-- DAT_I              Read data from slave. Result of a read transaction 
-- ERR_I              Slave interface error bus cycle acknowledgement signal. From slave to master. 
-- RTY_I              Slave interface retry bus cycle acknowledgement signal. From slave to master. - Not implemented
-- TGD_I              Data identification tag signal from slave.
--
-- Module dependent signals
-- WrtAddr            5-bit module select + 16-bit 'in module' address for write transactions
-- WrtData            32-bit data input from client logic for a Wishbone write transaction
-- WrtDataReq         Active high strobe from client logic to request a bus write transaction
-- WrtDataTag         see "Bus Data Tag identifiers" in implementation notes
-- WrtDataAck         Active high strobe from bus interface as write acknowledge.
-- RdAddr             5-bit module select + 16-bit 'in module' address valid for read transactions
-- RdData             32-bit data from interface to client logic as a result of a successful read transaction.
-- RdDataReq          Active high strobe from bus interface to indicate that RdData is valid
-- RdDataAck          Active high strobe from client logic to acknowledge the bus read transaction.   
-- RdDataType         see "Bus Data Tag identifiers" in implementation notes
-- SysClk             clock for client logic use. All transactions valid on rising edge of sysclk
-- Reset              reset to the client logic
--
-- Diagnostics        These registers are reset (cleared) by the RST_I signal.
-- BusEventStatus     Status register that latches bus error history. See description of BusEventStatus above. 
-- BusErrorReg        Counter - records the number of slave error acknowledgements.
-- BusGrantReg        Counter - records the number of times the master was refused a bus grant which resulted in a bus timeout
-- BusTimeOutReg      Counter - records the number of times a slave did not acknowledge a bus cycle
------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.ALL;
use ieee.std_logic_unsigned.ALL;
use ieee.numeric_std.ALL;

entity WbMasterInterfaceV106 is
   port ( CLK_I      : in    std_logic;
          RST_I      : in    std_logic;
          CYC_O      : out   std_logic; 
          CYC_GRNT   : in    std_logic; 
          STB_O      : out   std_logic_vector (7 downto 0); 
          WE_O       : out   std_logic; 
          ADR_O      : out   std_logic_vector (15 downto 0); 
          DAT_O      : out   std_logic_vector (31 downto 0); 
          TGD_O      : out   std_logic_vector(3 downto 0);
          TGC_O      : out   std_logic_vector(1 downto 0);
          ACK_I      : in    std_logic; 
          DAT_I      : in    std_logic_vector (31 downto 0); 
          ERR_I      : in    std_logic; 
          RTY_I      : in    std_logic; 
          TGD_I      : in    std_logic_vector(3 downto 0); 
          WrtAddr    : in    std_logic_vector (23 downto 0); 
          WrtData    : in    std_logic_vector (31 downto 0); 
          WrtDataReq : in    std_logic; 
          WrtDataTag : in    std_logic_vector(3 downto 0); 
          WrtDataAck : out   std_logic;
          RdAddr     : in    std_logic_vector (23 downto 0);
          RdData     : out   std_logic_vector (31 downto 0); 
          RdDataReq  : in    std_logic; 
          RdDataAck  : out   std_logic; 
          RdDataTag  : out   std_logic_vector(3 downto 0);
          StartExpReq    : in  std_logic;
          SysClk         : out   std_logic; 
          BusEventStatus : out std_logic_vector (3 downto 0);    
          BusErrorReg    : out std_logic_vector (3 downto 0);    
          BusGrantReg    : out std_logic_vector (3 downto 0);    
          BusTimeOutReg  : out std_logic_vector (3 downto 0)    
         );
end WbMasterInterfaceV106;

architecture BEHAVIORAL of WbMasterInterfaceV106 is
   constant BUSTIMEOUT : std_logic_vector(7 downto 0) := x"3C";     -- 60 bus clock cycles before bus transaction timeout occurs.

   type WrtMstrStateType is (WrtMstrIdle, WrtMstrArbitrate);
   signal WrtMstrState : WrtMstrStateType;
   signal WriteActive, WrtAckEnable  : std_logic;
   
   type RdMstrStateType is (RdMstrIdle, RdMstrArbitrate);
   signal RdMstrState : RdMstrStateType;
   signal ReadActive, RdAckEnable  : std_logic;
   
   signal SlaveAck : std_logic;
   signal BusWrtReq, WrtResetFlag : std_logic;

-- Diagnostics --
   signal BusTimerActive, BusTimeoutFlag : std_logic;
   signal BusTimer : std_logic_vector(7 downto 0);
   signal BusEventStatReg, BusErrorCnt, BusGrantCnt, BusTimeOutCnt : std_logic_vector(3 downto 0); 
begin

------------------------
-- Static assignments --
------------------------

-- Master Interface Input Signals
   SysClk <= CLK_I;                    -- System clock drives everything
   SlaveAck <= ACK_I or ERR_I or RTY_I or BusTimeoutFlag; 

-- Master Interface output Signals
   CYC_O <= WrtDataReq or RdDataReq or StartExpReq;
   WE_O  <= BusWrtReq;   
   WrtDataAck <= SlaveAck and WrtAckEnable and (WrtDataReq or StartExpReq);
   RdData <= DAT_I;
   RdDataTag <= TGD_I;
   RdDataAck  <= SlaveAck and RdAckEnable and RdDataReq;
   
-- Diagnostics
   BusEventStatus <= BusEventStatReg;
   BusErrorReg    <= BusErrorCnt;
   BusGrantReg    <= BusGrantCnt;
   BusTimeOutReg  <= BusTimeOutCnt;
   
----------------------------------
-- Interface => Bus Write logic --
----------------------------------   
   WBMasterWriteProc : process (CLK_I)
   begin
      if (rising_edge(CLK_I)) then
         if (RST_I = '1') then
            WrtAckEnable <= '0';
            WriteActive  <= '0';
            WrtMstrState <= WrtMstrIdle;            
         else
            case WrtMstrState is
               when WrtMstrIdle =>
                  if (ReadActive = '0' and (WrtDataReq = '1' or StartExpReq = '1') and WrtAckEnable = '0') then   -- Allow one cycle between repeated burst write resquests
                     WriteActive <= '1';
                     WrtMstrState <= WrtMstrArbitrate;                     
                  else
                     WrtAckEnable <= '0';
                     WriteActive <= '0';
                     WrtMstrState <= WrtMstrIdle;
                  end if;
               when WrtMstrArbitrate =>
                  WrtAckEnable <= '0';
                  if (CYC_GRNT = '1') then
                     WrtAckEnable <= '1';
                     if (SlaveAck = '1') then
                        if (WrtDataReq = '1') then
                           WrtMstrState <= WrtMstrArbitrate;
                        else
                           WrtMstrState <= WrtMstrIdle;
                        end if;
                     elsif (WrtDataReq = '1') then
                        WrtMstrState <= WrtMstrArbitrate;
                     else
                        WrtMstrState <= WrtMstrIdle;
                     end if;   
                  elsif (WrtDataReq = '1') then
                     WrtMstrState <= WrtMstrArbitrate;
                  else
                     WrtMstrState <= WrtMstrIdle;                  
                  end if;
               when others =>
                  WrtMstrState <= WrtMstrIdle;
            end case;
         end if;
      end if;
   end process;

-- Read cycle FSM --
   WBMasterReadProc : process (CLK_I)
   begin
      if (rising_edge(CLK_I)) then
         if (RST_I = '1') then
            RdMstrState <= RdMstrIdle;
            ReadActive <= '0';
            RdAckEnable <= '0';
         else
            case RdMstrState is
               when RdMstrIdle =>
                  if (WriteActive = '0' and RdDataReq = '1' and RdAckEnable = '0') then
                     ReadActive <= '1';
                     RdMstrState <= RdMstrArbitrate;
                  else
                     ReadActive <= '0';
                     RdAckEnable <= '0';
                     RdMstrState <= RdMstrIdle;
                  end if;
               when RdMstrArbitrate =>
                  RdAckEnable <= '0';
                  if (CYC_GRNT = '1') then
                     RdAckEnable <= '1';
                     if (SlaveAck = '1') then
                        if (RdDataReq = '1') then
                           RdMstrState <= RdMstrArbitrate;
                        else
                           RdMstrState <= RdMstrIdle;
                        end if;
                     elsif (RdDataReq = '1') then
                        RdMstrState <= RdMstrArbitrate;
                     else
                        RdMstrState <= RdMstrIdle;
                     end if;
                  elsif (RdDataReq = '1') then
                     RdMstrState <= RdMstrArbitrate;
                  else
                     RdMstrState <= RdMstrIdle;
                  end if;   
               when others =>
                  RdMstrState <= RdMstrIdle;
            end case;
         end if;
      end if;
   end process;

   WrtResetFlag <= '1' when (WrtAddr(15 downto 0) = x"FFFE") else '0';
   
-- Capture the address --
   WBMasterAddressProc : process (CLK_I)
   begin
      if (rising_edge(CLK_I)) then
         if (StartExpReq = '1' and ReadActive = '0') then -- Start exp request always takes precedence, except when read is currently active
            STB_O <= (others => '1');                    -- All modules are required to receive this
            ADR_O <= WrtAddr(15 downto 0);               -- The board address segment for writes which is the start exposure vector
            TGD_O <= WrtDataTag;
            TGC_O <= "10";                               -- Specific code for start exposure command            
            BusWrtReq  <= '1';
         elsif (WrtDataReq = '1' and ReadActive = '0') then -- Write request always takes next priority, again, except when read is currently active
            STB_O <= WrtAddr(23 downto 16);              -- The 'board select' segment for writes
            ADR_O <= WrtAddr(15 downto 0);               -- The board address segment for writes
            DAT_O <= WrtData;                            -- Capture the data
            TGD_O <= WrtDataTag;
            if (WrtResetFlag = '1') then
               TGC_O <= "11";
            else
               TGC_O <= "00";
            end if;
            BusWrtReq  <= '1';
         elsif (RdDataReq = '1') then
            STB_O <= RdAddr(23 downto 16);               -- The 'board select' segment for reads
            ADR_O <= RdAddr(15 downto 0);                -- The board address segment for reads
            TGD_O <= WrtDataTag;
            BusWrtReq  <= '0';
            TGC_O <= "00";
         else
            BusWrtReq  <= '0';          
            STB_O <= (others => '0');                    -- Kill the 'board select' strobes
         end if;
      end if;
   end process;

   WBMstrBusTimerProc : process (CLK_I)
   begin
      if (rising_edge(CLK_I)) then
         if (RST_I = '1') then
            BusTimeoutFlag <= '0';
            BusTimerActive <= '0';
            BusEventStatReg <= (others => '0');
            BusErrorCnt <= (others => '0');
            BusGrantCnt <= (others => '0');
            BusTimeOutCnt <= (others => '0');
         elsif (BusTimerActive = '0') then
            if (WrtDataReq = '1' or RdDataReq = '1' or StartExpReq = '1') then
               BusTimeoutFlag <= '0';
               BusTimerActive <= '1';
               BusTimer <= (others => '0');
            end if;
         elsif (SlaveAck = '1') then
            BusTimeoutFlag <= '0';
            BusTimerActive <= '0';
            if (ERR_I = '1') then
               BusEventStatReg <= BusEventStatReg or "0001";
               BusErrorCnt <= BusErrorCnt + 1;
            end if;
            if (RTY_I = '1') then
               BusEventStatReg <= BusEventStatReg or "0010";
            end if; 
         elsif (BusTimer >= BUSTIMEOUT) then
            BusTimeoutFlag <= '1';
            if (CYC_GRNT = '0') then
               BusEventStatReg <= BusEventStatReg or "1000";
               BusGrantCnt <= BusGrantCnt + 1;
            else
               BusEventStatReg <= BusEventStatReg or "0100";
               BusTimeOutCnt <= BusTimeOutCnt + 1;
            end if;
            BusTimerActive <= '0';
         else
            BusTimer <= BusTimer + 1;
         end if;
      end if;
   end process;

end BEHAVIORAL;


